# Copyright 2015, Yahoo Inc.
# Licensed under the terms of the Apache License, Version 2.0. See the LICENSE file associated with the project for terms.
from nose.tools import assert_true, assert_equal

import pickle as pkl
import sys
import os
import numpy as np
from sklearn.cross_validation import train_test_split

sys.path.insert(1, os.path.abspath('..'))
from lopq.model import LOPQModel, eigenvalue_allocation, accumulate_covariance_estimators, compute_rotations_from_accumulators
from lopq.search import LOPQSearcher, LOPQSearcherLMDB
from lopq.eval import compute_all_neighbors, get_cell_histogram, get_recall

########################################
# Helpers
########################################


relpath = lambda x: os.path.abspath(os.path.join(os.path.dirname(__file__), x))


def load_oxford_data():
    from lopq.utils import load_xvecs

    data = load_xvecs(relpath('../../data/oxford/oxford_features.fvecs'))
    return data


def make_random_model():
    m = LOPQModel(V=5, M=4, subquantizer_clusters=10)
    m.fit(np.random.RandomState(42).rand(200, 8), n_init=1)
    return m

########################################
# Tests
########################################


def test_eigenvalue_allocation():
    a = pkl.load(open(relpath('./testdata/test_eigenvalue_allocation_input.pkl')))

    vals, vecs = np.linalg.eigh(a)
    res = eigenvalue_allocation(4, vals)

    expected = np.array([
        63, 56, 52, 48, 44, 40, 36, 30, 26, 22, 18, 14, 10, 6, 3, 0,
        62, 57, 53, 51, 45, 41, 39, 33, 32, 31, 29, 25, 21, 17, 13, 9,
        61, 58, 54, 49, 47, 42, 38, 34, 28, 24, 20, 16, 12, 8, 5, 2,
        60, 59, 55, 50, 46, 43, 37, 35, 27, 23, 19, 15, 11, 7, 4, 1
    ])

    assert_true(np.equal(res, expected).all())


def test_eigenvalue_allocation_normalized_features():
    eigenvalues = np.array([
        2.02255824, 1.01940991, 0.01569471, 0.01355569, 0.01264379,
        0.01137654, 0.01108961, 0.01054673, 0.01023358, 0.00989679,
        0.00939045, 0.00900322, 0.00878857, 0.00870027, 0.00850136,
        0.00825236, 0.00813437, 0.00800231, 0.00790201, 0.00782219,
        0.00763405, 0.00752334, 0.00739174, 0.00728246, 0.00701366,
        0.00697365, 0.00677283, 0.00669658, 0.00654397, 0.00647679,
        0.00630645, 0.00621057
    ])
    indices = eigenvalue_allocation(2, eigenvalues)

    first_half = eigenvalues[indices[:16]]
    second_half = eigenvalues[indices[16:]]
    diff = np.abs(np.sum(np.log(first_half)) - np.sum(np.log(second_half)))
    assert_true(diff < .1, "eigenvalue_allocation is not working correctly")


def test_accumulate_covariance_estimators():
    data, centroids = pkl.load(open(relpath('./testdata/test_accumulate_covariance_estimators_input.pkl')))
    expected = pkl.load(open(relpath('./testdata/test_accumulate_covariance_estimators_output.pkl')))

    actual = accumulate_covariance_estimators(data, centroids)

    # Summed residual outer products
    assert_true(np.allclose(expected[0], actual[0]))

    # Summed residuals
    assert_true(np.allclose(expected[1], actual[1]))

    # Assignment count per cluster
    assert_true(np.array_equal(expected[2], actual[2]))

    # Assignments over data
    assert_true(np.array_equal(expected[3], actual[3]))

    # Residual data
    assert_true(np.allclose(expected[4], actual[4]))


def test_compute_rotations_from_accumulators():

    A, mu, count, num_buckets = pkl.load(open(relpath('./testdata/test_compute_rotations_from_accumulators_input.pkl')))
    expected = pkl.load(open(relpath('./testdata/test_compute_rotations_from_accumulators_output.pkl')))

    actual = compute_rotations_from_accumulators(A, mu, count, num_buckets)

    # Rotations
    assert_true(np.allclose(expected[0], actual[0]))

    # Mean residuals
    assert_true(np.allclose(expected[1], actual[1]))


def test_reconstruction():
    m = LOPQModel.load_proto(relpath('./testdata/random_test_model.lopq'))

    code = ((0, 1), (0, 1, 2, 3))
    r = m.reconstruct(code)
    expected = [-2.27444688, 6.47126941, 4.5042611, 4.76683476, 0.83671082, 9.36027283, 8.11780532, 6.34846377]

    assert_true(np.allclose(expected, r))


def test_oxford5k():

    random_state = 40
    data = load_oxford_data()
    train, test = train_test_split(data, test_size=0.2, random_state=random_state)

    # Compute distance-sorted neighbors in training set for each point in test set
    nns = compute_all_neighbors(test, train)

    # Fit model
    m = LOPQModel(V=16, M=8)
    m.fit(train, n_init=1, random_state=random_state)

    # Assert correct code computation
    assert_equal(m.predict(test[0]), ((3, 2), (14, 164, 83, 49, 185, 29, 196, 250)))

    # Assert low number of empty cells
    h = get_cell_histogram(train, m)
    assert_equal(np.count_nonzero(h == 0), 6)

    # Assert true NN recall on test set
    searcher = LOPQSearcher(m)
    searcher.add_data(train)
    recall, _ = get_recall(searcher, test, nns)
    assert_true(np.all(recall > [0.51, 0.92, 0.97, 0.97]))

    # Test partial fitting with just coarse quantizers
    m2 = LOPQModel(V=16, M=8, parameters=(m.Cs, None, None, None))
    m2.fit(train, n_init=1, random_state=random_state)

    searcher = LOPQSearcher(m2)
    searcher.add_data(train)
    recall, _ = get_recall(searcher, test, nns)
    assert_true(np.all(recall > [0.51, 0.92, 0.97, 0.97]))

    # Test partial fitting with coarse quantizers and rotations
    m3 = LOPQModel(V=16, M=8, parameters=(m.Cs, m.Rs, m.mus, None))
    m3.fit(train, n_init=1, random_state=random_state)

    searcher = LOPQSearcher(m3)
    searcher.add_data(train)
    recall, _ = get_recall(searcher, test, nns)
    assert_true(np.all(recall > [0.51, 0.92, 0.97, 0.97]))


def test_proto():
    import os

    filename = './temp_proto.lopq'
    m = make_random_model()
    m.export_proto(filename)
    m2 = LOPQModel.load_proto(filename)

    assert_equal(m.V, m2.V)
    assert_equal(m.M, m2.M)
    assert_equal(m.subquantizer_clusters, m2.subquantizer_clusters)

    assert_true(np.allclose(m.Cs[0], m2.Cs[0]))
    assert_true(np.allclose(m.Rs[0], m2.Rs[0]))
    assert_true(np.allclose(m.mus[0], m2.mus[0]))
    assert_true(np.allclose(m.subquantizers[0][0], m.subquantizers[0][0]))

    os.remove(filename)


def test_mat():
    import os

    filename = './temp_mat.mat'
    m = make_random_model()
    m.export_mat(filename)
    m2 = LOPQModel.load_mat(filename)

    assert_equal(m.V, m2.V)
    assert_equal(m.M, m2.M)
    assert_equal(m.subquantizer_clusters, m2.subquantizer_clusters)

    assert_true(np.allclose(m.Cs[0], m2.Cs[0]))
    assert_true(np.allclose(m.Rs[0], m2.Rs[0]))
    assert_true(np.allclose(m.mus[0], m2.mus[0]))
    assert_true(np.allclose(m.subquantizers[0][0], m.subquantizers[0][0]))

    os.remove(filename)


def searcher_instance_battery(searcher, q):
    """
    Helper to perform battery of searcher tests.
    """
    retrieved, visited = searcher.get_result_quota(q)
    assert_equal(len(retrieved), 12)
    assert_equal(visited, 3)

    retrieved, visited = searcher.search(q)
    assert_equal(len(retrieved), 10)
    assert_equal(visited, 3)

    retrieved, visited = searcher.get_result_quota(q, quota=20)
    assert_equal(len(retrieved), 28)
    assert_equal(visited, 5)

    retrieved, visited = searcher.search(q, quota=20)
    assert_equal(len(retrieved), 20)
    assert_equal(visited, 5)

    retrieved, visited = searcher.search(q, quota=20, limit=10)
    assert_equal(len(retrieved), 10)
    assert_equal(visited, 5)


def test_searcher():
    data = pkl.load(open(relpath('./testdata/test_searcher_data.pkl')))
    m = LOPQModel.load_proto(relpath('./testdata/random_test_model.lopq'))

    q = np.ones(8)

    # Test add_data
    searcher = LOPQSearcher(m)
    searcher.add_data(data)
    searcher_instance_battery(searcher, q)

    # Test add_codes
    searcher = LOPQSearcher(m)
    codes = [m.predict(x) for x in data]
    searcher.add_codes(codes)
    searcher_instance_battery(searcher, q)


def test_searcher_lmdb():
    import shutil

    data = pkl.load(open(relpath('./testdata/test_searcher_data.pkl')))
    m = LOPQModel.load_proto(relpath('./testdata/random_test_model.lopq'))

    lmbd_test_path = './test_lopq_lmbd'
    q = np.ones(8)

    # Test add_data
    searcher = LOPQSearcherLMDB(m, lmbd_test_path)
    searcher.add_data(data)
    searcher_instance_battery(searcher, q)

    # Clean up
    shutil.rmtree(lmbd_test_path)

    # Test add_codes
    searcher = LOPQSearcherLMDB(m, lmbd_test_path)
    codes = [m.predict(x) for x in data]
    searcher.add_codes(codes)
    searcher_instance_battery(searcher, q)

    # Clean up
    shutil.rmtree(lmbd_test_path)


def test_proto_partial():
    import os

    filename = './temp_proto_partial.lopq'
    c = (np.random.rand(8, 8), np.random.rand(8,8))
    m = LOPQModel(parameters=(c, None, None, None))
    m.export_proto(filename)
    m2 = LOPQModel.load_proto(filename)

    assert_equal(m.V, m2.V)
    assert_equal(m.M, m2.M)
    assert_equal(m.subquantizer_clusters, m2.subquantizer_clusters)

    assert_true(np.allclose(m.Cs[0], m2.Cs[0]))
    assert_true(m.Rs == m2.Rs)
    assert_true(m.mus == m2.mus)
    assert_true(m.subquantizers ==  m.subquantizers)

    os.remove(filename)



# Copyright 2015, Yahoo Inc.
# Licensed under the terms of the Apache License, Version 2.0. See the LICENSE file associated with the project for terms.
import heapq
from collections import defaultdict, namedtuple
from itertools import count
import numpy as np
import array
from .utils import iterate_splits, compute_codes_parallel


def multisequence(x, centroids):
    """
    Implementation of multi-sequence algorithm for traversing a multi-index.

    The algorithm is described in http://download.yandex.ru/company/cvpr2012.pdf.

    :param ndarray x:
        a query vector
    :param list centroids:
        a list of ndarrays containing cluster centroids for each subvector

    :yields int d:
        the cell distance approximation used to order cells
    :yields tuple cell:
        the cell indices
    """

    # Infer parameters
    splits = len(centroids)
    V = centroids[0].shape[0]

    # Compute distances to each coarse cluster and sort
    cluster_dists = []
    sorted_inds = []
    for cx, split in iterate_splits(x, splits):

        dists = ((cx - centroids[split]) ** 2).sum(axis=1)
        inds = np.argsort(dists)

        cluster_dists.append(dists)
        sorted_inds.append(inds)

    # Some helper functions used below
    def cell_for_inds(inds):
        return tuple([sorted_inds[s][i] for s, i in enumerate(inds)])

    def dist_for_cell(cell):
        return sum([cluster_dists[s][i] for s, i in enumerate(cell)])

    def inds_in_range(inds):
        for i in inds:
            if i >= V:
                return False
        return True

    # Initialize priority queue
    h = []
    traversed = set()
    start_inds = tuple(0 for _ in xrange(splits))
    start_dist = dist_for_cell(cell_for_inds(start_inds))
    heapq.heappush(h, (start_dist, start_inds))

    # Traverse cells
    while len(h):
        d, inds = heapq.heappop(h)
        yield d, cell_for_inds(inds)
        traversed.add(inds)

        # Add neighboring cells to queue
        if inds[1] == 0 or (inds[0] + 1, inds[1] - 1) in traversed:
            c = (inds[0] + 1, inds[1])
            if inds_in_range(c):
                dist = dist_for_cell(cell_for_inds(c))
                heapq.heappush(h, (dist, c))

        if inds[0] == 0 or (inds[0] - 1, inds[1] + 1) in traversed:
            c = (inds[0], inds[1] + 1)
            if inds_in_range(c):
                dist = dist_for_cell(cell_for_inds(c))
                heapq.heappush(h, (dist, c))


class LOPQSearcherBase(object):

    def add_data(self, data, ids=None, num_procs=1):
        """
        Add raw data into the search index.

        :param ndarray data:
            an ndarray with data points on the rows
        :param ndarray ids:
            an optional array of ids for each data point;
            defaults to the index of the data point if not provided
        :param int num_procs:
            an integer specifying the number of processes to use to
            compute codes for the data
        """
        codes = compute_codes_parallel(data, self.model, num_procs)
        self.add_codes(codes, ids)

    def get_result_quota(self, x, quota=10):
        """
        Given a query vector and result quota, retrieve as many cells as necessary
        to fill the quota.

        :param ndarray x:
            a query vector
        :param int quota:
            the desired number of items to retrieve

        :returns list retrieved:
            a list of index items
        :returns int visited:
            the number of multi-index cells visited
        """
        retrieved = []
        visited = 0
        for _, cell in multisequence(x, self.model.Cs):
            retrieved += self.get_cell(cell)
            visited += 1

            if len(retrieved) >= quota:
                break

        return retrieved, visited

    def compute_distances(self, x, items):
        """
        Given a query and a list of index items, compute the approximate distance of the query
        to each item and return a list of tuples that contain the distance and the item.
        Memoize subquantizer distances per coarse cluster to save work.

        :param ndarray x:
            a query vector
        :param list items:
            a list of items from the index

        :returns list:
            a list of items with distance
        """
        memoized_subquant_dists = [{}, {}]

        def get_subquantizer_distances(x, coarse):

            d0, d1 = memoized_subquant_dists
            c0, c1 = coarse

            if c0 not in d0:
                d0[c0] = self.model.get_subquantizer_distances(x, coarse, coarse_split=0)

            if c1 not in d1:
                d1[c1] = self.model.get_subquantizer_distances(x, coarse, coarse_split=1)

            return d0[c0] + d1[c1]

        results = []
        for item in items:

            codes = item[1]
            coarse, fine = codes

            subquantizer_distances = get_subquantizer_distances(x, coarse)
            dist = sum([subquantizer_distances[i][fc] for i, fc in enumerate(fine)])

            results.append((dist, item))

        return results

    def search(self, x, quota=10, limit=None, with_dists=False):
        """
        Return euclidean distance ranked results, along with the number of cells
        traversed to fill the quota.

        :param ndarray x:
            a query vector
        :param int quota:
            the number of desired results to rank
        :param int limit:
            the number of desired results to return - defaults to quota
        :param bool with_dists:
            boolean indicating whether result items should be returned with their distance

        :returns list results:
            the list of ranked results
        :returns int visited:
            the number of cells visited in the query
        """
        # Retrieve results with multi-index
        retrieved, visited = self.get_result_quota(x, quota)

        # Compute distance for results
        results = self.compute_distances(x, retrieved)

        # Sort by distance
        results = sorted(results, key=lambda d: d[0])

        # Limit number returned
        if limit is None:
            limit = quota
        results = results[:limit]

        if with_dists:
            Result = namedtuple('Result', ['id', 'code', 'dist'])
            results = map(lambda d: Result(d[1][0], d[1][1], d[0]), results)
        else:
            Result = namedtuple('Result', ['id', 'code'])
            results = map(lambda d: Result(d[1][0], d[1]), results)

        return results, visited

    def add_codes(self, codes, ids=None):
        """
        Add LOPQ codes into the search index.

        :param iterable codes:
            an iterable of LOPQ code tuples
        :param iterable ids:
            an optional iterable of ids for each code;
            defaults to the index of the code tuple if not provided
        """
        raise NotImplementedError()

    def get_cell(self, cell):
        """
        Retrieve a cell bucket from the index.

        :param tuple cell:
            a cell tuple

        :returns list:
            the list of index items in this cell bucket
        """
        raise NotImplementedError()

class LOPQSearcher(LOPQSearcherBase):
    def __init__(self, model):
        """
        Create an LOPQSearcher instance that encapsulates retrieving and ranking
        with LOPQ. Requires an LOPQModel instance. This class uses a Python dict
        to implement the index.

        :param LOPQModel model:
            the model for indexing and ranking
        """
        self.model = model
        self.index = defaultdict(list)

    def add_codes(self, codes, ids=None):
        """
        Add LOPQ codes into the search index.

        :param iterable codes:
            an iterable of LOPQ code tuples
        :param iterable ids:
            an optional iterable of ids for each code;
            defaults to the index of the code tuple if not provided
        """
        # If a list of ids is not provided, assume it is the index of the data
        if ids is None:
            ids = count()

        for item_id, code in zip(ids, codes):
            cell = code[0]
            self.index[cell].append((item_id, code))

    def get_cell(self, cell):
        """
        Retrieve a cell bucket from the index.

        :param tuple cell:
            a cell tuple

        :returns list:
            the list of index items in this cell bucket
        """
        return self.index[cell]


class LOPQSearcherLMDB(LOPQSearcherBase):
    def __init__(self, model, lmdb_path, id_lambda=int):
        """
        Create an LOPQSearcher instance that encapsulates retrieving and ranking
        with LOPQ. Requires an LOPQModel instance. This class uses an lmbd database
        to implement the index.

        :param LOPQModel model:
            the model for indexing and ranking
        :param str lmdb_path:
            path for the lmdb database; if it does not exist it is created
        :param callable id_lambda:
            a lambda function to reconstruct item ids from their string representation
            (computed by calling `bytes`) during retrieval
        """
        import lmdb

        self.model = model
        self.lmdb_path = lmdb_path
        self.id_lambda = id_lambda

        self.env = lmdb.open(self.lmdb_path, map_size=1024*2000000*2, writemap=False, map_async=True, max_dbs=1)
        self.index_db = self.env.open_db("index")

    def encode_cell(self, cell):
        return array.array("H", cell).tostring()

    def decode_cell(self, cell_bytes):
        a = array.array("H")
        a.fromstring(cell_bytes)
        return tuple(a.tolist())

    def encode_fine_codes(self, fine):
        return array.array("B", fine).tostring()

    def decode_fine_codes(self, fine_bytes):
        a = array.array("B")
        a.fromstring(fine_bytes)
        return tuple(a.tolist())

    def add_codes(self, codes, ids=None):
        """
        Add LOPQ codes into the search index.

        :param iterable codes:
            an iterable of LOPQ code tuples
        :param iterable ids:
            an optional iterable of ids for each code;
            defaults to the index of the code tuple if not provided
        """
        # If a list of ids is not provided, assume it is the index of the data
        if ids is None:
            ids = count()

        with self.env.begin(db=self.index_db, write=True) as txn:
            for item_id, code in zip(ids, codes):
                key_prefix = self.encode_cell(code[0])
                key_suffix = bytes(item_id)
                key = key_prefix + key_suffix
                val = self.encode_fine_codes(code[1])
                txn.put(key, val)
        self.env.sync()

    def get_cell(self, cell):
        """
        Retrieve a cell bucket from the index.

        :param tuple cell:
            a cell tuple

        :returns list:
            the list of index items in this cell bucket
        """
        prefix = self.encode_cell(cell)

        items = []
        with self.env.begin(db=self.index_db) as txn:
            cursor = txn.cursor()
            cursor.set_range(prefix)
            for key, value in cursor:
                if not key.startswith(prefix):
                    break
                else:
                    item_id = self.id_lambda(key[4:])
                    cell = self.decode_cell(key[:4])
                    fine = self.decode_fine_codes(value)
                    code = (cell, fine)
                    items.append((item_id, code))
            cursor.close()

        return items



# Copyright 2015, Yahoo Inc.
# Licensed under the terms of the Apache License, Version 2.0. See the LICENSE file associated with the project for terms.
import numpy as np
from sklearn.cluster import KMeans
import logging
import sys
from collections import namedtuple
from .utils import iterate_splits, predict_cluster

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)
logger.addHandler(logging.StreamHandler(sys.stdout))


########################################
# Core training algo
########################################

def eigenvalue_allocation(num_buckets, eigenvalues):
    """
    Compute a permutation of eigenvalues to balance variance accross buckets
    of dimensions.

    Described in section 3.2.4 in http://research.microsoft.com/pubs/187499/cvpr13opq.pdf

    Note, the following slides indicate this function will break when fed eigenvalues < 1
    without the scaling trick implemented below:

        https://www.robots.ox.ac.uk/~vgg/rg/slides/ge__cvpr2013__optimizedpq.pdf


    :param int num_buckets:
        the number of dimension buckets over which to allocate eigenvalues
    :param ndarray eigenvalues:
        a vector of eigenvalues

    :returns ndarray:
        a vector of indices by which to permute the eigenvectors
    """
    D = len(eigenvalues)
    dims_per_bucket = D / num_buckets
    eigenvalue_product = np.zeros(num_buckets, dtype=float)
    bucket_size = np.zeros(num_buckets, dtype=int)
    permutation = np.zeros((num_buckets, dims_per_bucket), dtype=int)

    # We first must scale the eigenvalues by dividing by their
    # smallets non-zero value to avoid problems with the algorithm
    # when eigenvalues are less than 1.
    min_non_zero_eigenvalue = np.min(np.abs(eigenvalues[np.nonzero(eigenvalues)]))
    eigenvalues = eigenvalues / min_non_zero_eigenvalue

    # Iterate eigenvalues in descending order
    sorted_inds = np.argsort(eigenvalues)[::-1]
    log_eigs = np.log2(abs(eigenvalues))
    for ind in sorted_inds:

        # Find eligible (not full) buckets
        eligible = (bucket_size < dims_per_bucket).nonzero()

        # Find eligible bucket with least eigenvalue product
        i = eigenvalue_product[eligible].argmin(0)
        bucket = eligible[0][i]

        # Update eigenvalue product for this bucket
        eigenvalue_product[bucket] = eigenvalue_product[bucket] + log_eigs[ind]

        # Store bucket assignment and update size
        permutation[bucket, bucket_size[bucket]] = ind
        bucket_size[bucket] += 1

    return np.reshape(permutation, D)


def compute_local_rotations(data, C, num_buckets):
    """
    Compute a rotation matrix for each cluster in the model by estimating
    the covariance matrix for each cluster and balancing variance across
    buckets of the dimensions.

    :param ndarray data:
        data points, each row an observation
    :param ndarray C:
        a VxD matrix of cluster centroids
    :param int num_buckets:
        the number of buckets accross which to balance variance in the rotation matrix

    :returns ndarray:
        a VxDxD tensor containing the rotation matrix for each cluster
        where V is the number of clusters and D the dimension of the data (which is split here)
    :returns ndarray:
        a Nx1 vector of cluster assignments for each data point
    :returns ndarray:
        a VxD vector of mean residuals for each cluster
    :returns ndarray:
        an NxD matrix of residuals
    """

    logger.info('Fitting local rotations...')

    A, mu, count, assignments, residuals = accumulate_covariance_estimators(data, C)

    R, mu = compute_rotations_from_accumulators(A, mu, count, num_buckets)

    logger.info('Done fitting local rotations.')

    return R, mu, assignments, residuals


def accumulate_covariance_estimators(data, C):
    """
    Accumulate covariance estimators for each cluster with a pass through the data.

    :param ndarray data:
        NxD array - observations on the rows
    :param ndarray C:
        VxD array of cluster centroids

    :returns ndarray A:
        VxDxD array - total sum of residual outer products for each cluster
    :returns ndarray mu:
        VxD array of total sum of residuals per cluster
    :returns ndarray count:
        Vx1 array of cluster sizes
    :returns ndarray assignments:
        Nx1 array of cluster assignments
    :returns ndarray residuals:
        NxD array of data residuals
    """

    V = C.shape[0]
    N = data.shape[0]
    D = data.shape[1]

    # Essential variables
    A = np.zeros((V, D, D))                 # accumulators for covariance estimator per cluster
    mu = np.zeros((V, D))                   # residual means
    count = np.zeros(V, dtype=int)          # count of points per cluster
    assignments = np.zeros(N, dtype=int)    # point cluster assignments
    residuals = np.zeros((N, D))            # residual for data points given cluster assignment

    # Iterate data points, accumulate estimators
    for i in xrange(N):
        d = data[i]

        # Find cluster assignment and residual
        cluster = predict_cluster(d, C)
        centroid = C[cluster]
        residual = d - centroid
        assignments[i] = cluster

        # Accumulate estimators for covariance matrix for the assigned cluster
        mu[cluster] += residual
        count[cluster] += 1
        A[cluster] += np.outer(residual, residual)
        residuals[i] = residual

    return A, mu, count, assignments, residuals


def compute_rotations_from_accumulators(A, mu, count, num_buckets):
    """
    Given accumulators computed on cluster residuals, compute the optimal
    rotation matrix. The A and mu variables are modified in place and returned to
    avoid memory allocation.

    :param ndarray A:
        a VxDxD array - total sum of outer products of residuals per cluster
    :param ndarray mu:
        a VxD array - total sum of residuals per cluster
    :param ndarray count:
        a Vx1 array - count of points for each cluster
    :param int num_buckets:
        the number of subvectors to balance variance across

    :returns ndarray A:
        a VxDxD array - per cluster local rotations of size DxD on the first dimension
    :returns ndarray mu:
        a VxD array of mean residuals per cluster
    """

    V, D = mu.shape

    # For each cluster, use accumulator variables to estimate covariance matrix
    # and compute rotation matrix
    for i in xrange(V):

        # Normalize
        num_points = count[i]
        mu[i] /= num_points

        # Compute covariance estimator
        cov = (A[i] + A[i].transpose()) / (2 * (num_points - 1)) - np.outer(mu[i], mu[i])

        # Compute eigenvalues, reuse A matrix
        if num_points < D:
            logger.warn('Fewer points (%d) than dimensions (%d) in rotation computation for cluster %d' % (num_points, D, i))
            eigenvalues = np.ones(D)
            A[i] = np.eye(D)
        else:
            eigenvalues, A[i] = np.linalg.eigh(cov)

        # Permute eigenvectors to balance variance in subquantizers
        permuted_inds = eigenvalue_allocation(num_buckets, eigenvalues)
        A[i] = A[i, :, permuted_inds]

    return A, mu


def project_residuals_to_local(residuals, assignments, Rs, mu):
    """
    Given residuals for training datapoints, their cluster assignments,
    the mean of cluster residuals, and the cluster rotation matrices, project
    all residuals to their appropriate local frame. This is run to generate
    points to cluster for subquantizer training.

    :params ndarray residuals:
        an NxD array of residuals
    :params ndarray assignments:
        an Nx1 array of cluster ids assignments for residuals
    :params ndarray Rs:
        a VxDxD array of rotation matrices for each cluster
    :params ndarray mu:
        a VxD matrix of mean residuals for each cluster

    :returns ndarray:
        an NxD array of locally projected residuals
    """
    projected = np.zeros(residuals.shape)
    for i in xrange(residuals.shape[0]):
        res = residuals[i]
        a = assignments[i]
        projected[i] = np.dot(Rs[a], res - mu[a])

    return projected


def compute_residuals(data, C):
    assignments = np.apply_along_axis(predict_cluster, 1, data, C)
    residuals = data - C[assignments]
    return residuals, assignments


def train_coarse(data, V=8, kmeans_coarse_iters=10, n_init=10, random_state=None):
    """
    Train a kmeans model.

    :param ndarray data:
        an NxD array with observations on the rows
    :param int V:
        the number of clusters
    :param int kmeans_coarse_iters:
        the nubmer of iterations
    :param int random_state:
        a random state to seed the clustering

    :returns ndarray:
        a VxD matrix of cluster centroids
    """

    logger.info('Fitting coarse quantizer...')

    # Fit coarse model
    model = KMeans(n_clusters=V, init="k-means++", max_iter=kmeans_coarse_iters, n_init=n_init, n_jobs=1, verbose=False, random_state=random_state)
    model.fit(data)

    logger.info('Done fitting coarse quantizer.')

    return model.cluster_centers_


def train_subquantizers(data, num_buckets, subquantizer_clusters=256, kmeans_local_iters=20, n_init=10, random_state=None):
    """
    Fit a set of num_buckets subquantizers for corresponding subvectors.
    """

    subquantizers = list()
    for i, d in enumerate(np.split(data, num_buckets, axis=1)):
        model = KMeans(n_clusters=subquantizer_clusters, init="k-means++", max_iter=kmeans_local_iters,
                       n_init=n_init, n_jobs=1, verbose=False, random_state=random_state)
        model.fit(d)
        subquantizers.append(model.cluster_centers_)
        logger.info('Fit subquantizer %d of %d.' % (i + 1, num_buckets))

    return subquantizers


def train(data, V=8, M=4, subquantizer_clusters=256, parameters=None,
          kmeans_coarse_iters=10, kmeans_local_iters=20, n_init=10,
          subquantizer_sample_ratio=1.0, random_state=None, verbose=False):
    """
    Fit an LOPQ model.

    :param ndarray data:
        a NxD matrix of training data points with observations on the rows

    :param int V:
        number of coarse clusters
    :param int M:
        number of fine codes; same as number of bytes per compressed
        vector in memory with 256 subquantizer clusters
    :param int subquantizer_clusters:
        the number of clusters for each subquantizer
    :param tuple parameters:
        a tuple of parameters - missing parameters are allowed to be None

    :param int kmeans_coarse_iters:
        kmeans iterations
    :param int kmeans_local_iters:
        kmeans iterations for subquantizers
    :param int n_init:
        the number of independent kmeans runs for all kmeans when training - set low for faster training
    :param float subquantizer_sample_ratio:
        the proportion of the training data to sample for training subquantizers - since the number of
        subquantizer clusters is much smaller then the number of coarse clusters, less data is needed
    :param int random_state:
        a random seed used in all random operations during training if provided
    :param bool verbose:
        a bool enabling verbose output during training

    :returns tuple:
        a tuple of model parameters that can be used to instantiate an LOPQModel object
    """

    # Set logging level for verbose mode
    if (verbose):
        logger.setLevel(logging.DEBUG)

    # Extract parameters
    Cs = Rs = mus = subquantizers = None
    if parameters is not None:
        Cs, Rs, mus, subquantizers = parameters

    # Enforce parameter dependencies
    if Rs is None or mus is None:
        Rs = mus = None

    # Split vectors
    # TODO: permute dims here if this hasn't already been done
    first_half, second_half = np.split(data, 2, axis=1)

    # Cluster coarse splits
    if Cs is not None:
        logger.info('Using existing coarse quantizers.')
        C1, C2 = Cs
    else:
        C1 = train_coarse(first_half, V, kmeans_coarse_iters, n_init, random_state)
        C2 = train_coarse(second_half, V, kmeans_coarse_iters, n_init, random_state)

    # Compute local rotations
    if Rs is not None and mus is not None:
        logger.info('Using existing rotations.')
        Rs1, Rs2 = Rs
        mu1, mu2 = mus
        assignments1 = assignments2 = residuals1 = residuals2 = None
    else:
        Rs1, mu1, assignments1, residuals1 = compute_local_rotations(first_half, C1, M / 2)
        Rs2, mu2, assignments2, residuals2 = compute_local_rotations(second_half, C2, M / 2)

    # Subquantizers don't need as much data, so we could sample here
    subquantizer_sample_ratio = min(subquantizer_sample_ratio, 1.0)
    N = data.shape[0]
    N2 = int(np.floor(subquantizer_sample_ratio * N))
    sample_inds = np.random.RandomState(random_state).choice(N, N2, False)
    logger.info('Sampled training data for subquantizers with %f proportion (%d points).' % (subquantizer_sample_ratio, N2))

    # Use assignments and residuals from rotation computation if available
    if assignments1 is not None:
        residuals1 = residuals1[sample_inds]
        residuals2 = residuals2[sample_inds]
        assignments1 = assignments1[sample_inds]
        assignments2 = assignments2[sample_inds]
    else:
        residuals1, assignments1 = compute_residuals(first_half[sample_inds], C1)
        residuals2, assignments2 = compute_residuals(second_half[sample_inds], C2)

    # Project residuals
    logger.info('Projecting residuals to local frame...')
    projected1 = project_residuals_to_local(residuals1, assignments1, Rs1, mu1)
    projected2 = project_residuals_to_local(residuals2, assignments2, Rs2, mu2)

    logger.info('Fitting subquantizers...')
    subquantizers1 = train_subquantizers(projected1, M / 2, subquantizer_clusters, kmeans_local_iters, n_init, random_state=random_state)
    subquantizers2 = train_subquantizers(projected2, M / 2, subquantizer_clusters, kmeans_local_iters, n_init, random_state=random_state)
    logger.info('Done fitting subquantizers.')

    return (C1, C2), (Rs1, Rs2), (mu1, mu2), (subquantizers1, subquantizers2)

########################################
# Model class
########################################

# Named tuple type for LOH codes
LOPQCode = namedtuple('LOPQCode', ['coarse', 'fine'])


class LOPQModel(object):
    def __init__(self, V=8, M=4, subquantizer_clusters=256, parameters=None):
        """
        Create an LOPQModel instance that encapsulates a complete LOPQ model with parameters and hyperparameters.

        :param int V:
            the number of clusters per a coarse split
        :param int M:
            the total number of subvectors (equivalent to the total number of subquantizers)
        :param int subquantizer_clusters:
            the number of clusters for each subquantizer
        :param tuple parameters:
            a tuple of parameters - missing parameters are allowed to be None

            the tuple will look like the following

            ((C1, C2), (Rs1, Rs2), (mu1, mu2), (subquantizers1, subquantizers2))

            where each element is itself a pair with one split of parameters for the each of the coarse splits.

            the parameters have the following data types (V and M have the meaning described above,
            D is the total dimension of the data, and S is the number of subquantizer clusters):

                C: VxD/2 ndarray of coarse centroids
                R: VxD/2xD/2 ndarray of fitted rotation matrices for each coarse cluster
                mu: VxD/2 ndarray of mean residuals for each coar cluster
                subquantizer: length M/2 list of SxD/M ndarrays of cluster centroids for each subvector
        """

        # If learned parameters are passed in explicitly, derive the model params by inspection.
        self.Cs, self.Rs, self.mus, self.subquantizers = parameters if parameters is not None else (None, None, None, None)

        if self.Cs is not None:
            self.V = self.Cs[0].shape[0]
            self.num_coarse_splits = len(self.Cs)
        else:
            self.V = V
            self.num_coarse_splits = 2

        if self.subquantizers is not None:
            self.num_fine_splits = len(self.subquantizers[0])
            self.M = self.num_fine_splits * self.num_coarse_splits
            self.subquantizer_clusters = self.subquantizers[0][0].shape[0]
        else:
            self.num_fine_splits = M / 2
            self.M = M
            self.subquantizer_clusters = subquantizer_clusters

    def fit(self, data, kmeans_coarse_iters=10, kmeans_local_iters=20, n_init=10, subquantizer_sample_ratio=1.0, random_state=None, verbose=False):
        """
        Fit a model with the current model parameters. This method will use existing parameters and only
        train missing parameters.

        :param int kmeans_coarse_iters:
            the number of kmeans iterations for coarse quantizer training
        :param int kmeans_local_iters:
            the number of kmeans iterations for subquantizer taining
        :param int n_init:
            the number of independent kmeans runs for all kmeans when training - set low for faster training
        :param float subquantizer_sample_ratio:
            the proportion of the training data to sample for training subquantizers - since the number of
            subquantizer clusters is much smaller then the number of coarse clusters, less data is needed
        :param int random_state:
            a random seed used in all random operations during training if provided
        :param bool verbose:
            a bool enabling verbose output during training
        """
        existing_parameters = (self.Cs, self.Rs, self.mus, self.subquantizers)

        parameters = train(data, self.V, self.M, self.subquantizer_clusters, existing_parameters,
                           kmeans_coarse_iters, kmeans_local_iters, n_init, subquantizer_sample_ratio,
                           random_state, verbose)

        self.Cs, self.Rs, self.mus, self.subquantizers = parameters

    def get_split_parameters(self, split):
        """
        A helper to return parameters for a given coarse split.

        :params int split:
            the coarse split

        :returns ndarray:
            a matrix of centroids for the coarse model
        :returns list:
            a list of residual means for each cluster
        :returns list:
            a list of rotation matrices for each cluster
        :returns list:
            a list of centroid matrices for each subquantizer in this coarse split
        """
        return self.Cs[split] if self.Cs is not None else None, \
            self.Rs[split] if self.Rs is not None else None, \
            self.mus[split] if self.mus is not None else None, \
            self.subquantizers[split] if self.subquantizers is not None else None

    def predict(self, x):
        """
        Compute both coarse and fine codes for a datapoint.

        :param ndarray x:
            the point to code

        :returns tuple:
            a tuple of coarse codes
        :returns tuple:
            a tuple of fine codes
        """
        # Compute coarse quantizer codes
        coarse_codes = self.predict_coarse(x)

        # Compute fine codes
        fine_codes = self.predict_fine(x, coarse_codes)

        return LOPQCode(coarse_codes, fine_codes)

    def predict_coarse(self, x):
        """
        Compute the coarse codes for a datapoint.

        :param ndarray x:
            the point to code

        :returns tuple:
            a tuple of coarse codes
        """
        return tuple([predict_cluster(cx, self.Cs[split]) for cx, split in iterate_splits(x, self.num_coarse_splits)])

    def predict_fine(self, x, coarse_codes=None):
        """
        Compute the fine codes for a datapoint.

        :param ndarray x:
            the point to code
        :param ndarray coarse_codes:
            the coarse codes for the point
            if they are already computed

        :returns tuple:
            a tuple of fine codes
        """
        if coarse_codes is None:
            coarse_codes = self.predict_coarse(x)

        px = self.project(x, coarse_codes)

        fine_codes = []
        for cx, split in iterate_splits(px, self.num_coarse_splits):

            # Get product quantizer parameters for this split
            _, _, _, subC = self.get_split_parameters(split)

            # Compute subquantizer codes
            fine_codes += [predict_cluster(fx, subC[sub_split]) for fx, sub_split in iterate_splits(cx, self.num_fine_splits)]

        return tuple(fine_codes)

    def project(self, x, coarse_codes, coarse_split=None):
        """
        Project this vector to its local residual space defined by the coarse codes.

        :param ndarray x:
            the point to project
        :param ndarray coarse_codes:
            the coarse codes defining the local space
        :param int coarse_split:
            index of the coarse split to get distances for - if None then all splits
            are computed

        :returns ndarray:
            the projected vector
        """
        px = []

        if coarse_split is None:
            split_iter = iterate_splits(x, self.num_coarse_splits)
        else:
            split_iter = [(np.split(x, self.num_coarse_splits)[coarse_split], coarse_split)]

        for cx, split in split_iter:

            # Get product quantizer parameters for this split
            C, R, mu, _ = self.get_split_parameters(split)

            # Retrieve already computed coarse cluster
            cluster = coarse_codes[split]

            # Compute residual
            r = cx - C[cluster]

            # Project residual to local frame
            pr = np.dot(R[cluster], r - mu[cluster])
            px.append(pr)

        return np.concatenate(px)

    def reconstruct(self, codes):
        """
        Given a code tuple, reconstruct an approximate vector.

        :param tuple codes:
            a code tuple as returned from the predict method

        :returns ndarray:
            a reconstructed vector
        """
        coarse_codes, fine_codes = codes

        x = []
        for fc, split in iterate_splits(fine_codes, self.num_coarse_splits):

            # Get product quantizer parameters for this split
            C, R, mu, subC = self.get_split_parameters(split)

            # Concatenate the cluster centroids for this split of fine codes
            sx = reduce(lambda acc, c: np.concatenate((acc, subC[c[0]][c[1]])), enumerate(fc), [])

            # Project residual out of local space
            cluster = coarse_codes[split]
            r = np.dot(R[cluster].transpose(), sx) + mu[cluster]

            # Reconstruct from cluster centroid
            x = np.concatenate((x, r + C[cluster]))

        return x

    def get_subquantizer_distances(self, x, coarse_codes, coarse_split=None):
        """
        Project a given query vector to the local space of the given coarse codes
        and compute the distances of each subvector to the corresponding subquantizer
        clusters.

        :param ndarray x:
            a query  vector
        :param tuple coarse_codes:
            the coarse codes defining which local space to project to
        :param int coarse_split:
            index of the coarse split to get distances for - if None then all splits
            are computed

        :returns list:
            a list of distances to each subquantizer cluster for each subquantizer
        """

        px = self.project(x, coarse_codes)
        subquantizer_dists = []

        if coarse_split is None:
            split_iter = iterate_splits(px, self.num_coarse_splits)
        else:
            split_iter = [(np.split(px, self.num_coarse_splits)[coarse_split], coarse_split)]

        # for cx, split in iterate_splits(px, self.num_coarse_splits):
        for cx, split in split_iter:
            _, _, _, subC = self.get_split_parameters(split)
            subquantizer_dists += [((fx - subC[sub_split]) ** 2).sum(axis=1) for fx, sub_split in iterate_splits(cx, self.num_fine_splits)]

        return subquantizer_dists

    def get_cell_id_for_coarse_codes(self, coarse_codes):
        return coarse_codes[1] + coarse_codes[0] * self.V

    def get_coarse_codes_for_cell_id(self, cell_id):
        return (int(np.floor(float(cell_id) / self.V)), cell_id % self.V)

    def export_mat(self, filename):
        """
        Export model parameters in .mat file format.

        Splits in the parameters (coarse splits and fine splits) are concatenated together in the
        resulting arrays. For example, the Cs paramaters become a 2 x V x D array where the first dimension
        indexes the split. The subquantizer centroids are encoded similarly as a 2 x (M/2) x 256 x (D/M) array.
        """
        from scipy.io import savemat
        from .utils import concat_new_first

        Cs = concat_new_first(self.Cs)
        Rs = concat_new_first(self.Rs)
        mus = concat_new_first(self.mus)
        subs = concat_new_first(map(concat_new_first, self.subquantizers))

        savemat(filename, {'Cs': Cs, 'Rs': Rs, 'mus': mus, 'subs': subs, 'V': self.V, 'M': self.M})

    @staticmethod
    def load_mat(filename):
        """
        Reconstitute an LOPQModel in the format exported by the `export_mat` method above.
        """
        from scipy.io import loadmat

        d = loadmat(filename)

        M = d['M'][0][0]
        Cs = tuple(map(np.squeeze, np.split(d['Cs'], 2, axis=0)))
        Rs = tuple(map(np.squeeze, np.split(d['Rs'], 2, axis=0)))
        mus = tuple(map(np.squeeze, np.split(d['mus'], 2, axis=0)))

        subs = tuple([map(np.squeeze, np.split(half, M / 2, axis=0)) for half in map(np.squeeze, np.split(d['subs'], 2, axis=0))])

        return LOPQModel(parameters=(Cs, Rs, mus, subs))

    def export_proto(self, f):
        """
        Export model parameters in protobuf format.
        """
        from .lopq_model_pb2 import LOPQModelParams
        from itertools import chain

        lopq_params = LOPQModelParams()
        lopq_params.D = 2 * self.Cs[0].shape[1]
        lopq_params.V = self.V
        lopq_params.M = self.M
        lopq_params.num_subquantizers = self.subquantizer_clusters

        def matrix_from_ndarray(m, a):
            m.values.extend(map(float, np.nditer(a, order='C')))
            m.shape.extend(a.shape)
            return m

        def vector_from_ndarray(m, a):
            m.values.extend(map(float, np.nditer(a, order='C')))
            return m

        if self.Cs != None:
            for C in self.Cs:
                matrix_from_ndarray(lopq_params.Cs.add(), C)
        if self.Rs != None:
            for R in chain(*self.Rs):
                matrix_from_ndarray(lopq_params.Rs.add(), R)
        if self.mus != None:
            for mu in chain(*self.mus):
                vector_from_ndarray(lopq_params.mus.add(), mu)
        if self.subquantizers != None:
            for sub in chain(*self.subquantizers):
                matrix_from_ndarray(lopq_params.subs.add(), sub)

        if type(f) is str:
            f = open(f, 'wb')
        f.write(lopq_params.SerializeToString())
        f.close()

    @staticmethod
    def load_proto(filename):
        """
        Reconstitute a model from parameters stored in protobuf format.
        """
        from .lopq_model_pb2 import LOPQModelParams
        from .utils import concat_new_first

        def halves(arr):
            return [arr[:len(arr) / 2], arr[len(arr) / 2:]]

        lopq_params = LOPQModelParams()

        try:
            f = open(filename)
            lopq_params.ParseFromString(f.read())
            f.close()

            Cs = Rs = mus = subs = None
            if len(lopq_params.Cs) != 0:
                Cs = [np.reshape(C.values, C.shape) for C in lopq_params.Cs]
            if len(lopq_params.Rs) != 0:
                Rs = map(concat_new_first, halves([np.reshape(R.values, R.shape) for R in lopq_params.Rs]))
            if len(lopq_params.mus) != 0:
                mus = map(concat_new_first, halves([np.array(mu.values) for mu in lopq_params.mus]))
            if len(lopq_params.subs) != 0:
                subs = halves([np.reshape(sub.values, sub.shape) for sub in lopq_params.subs])

            return LOPQModel(parameters=(Cs, Rs, mus, subs))

        except IOError:
            print filename + ": Could not open file."
            return None



# Copyright 2015, Yahoo Inc.
# Licensed under the terms of the Apache License, Version 2.0. See the LICENSE file associated with the project for terms.
import time
import numpy as np


def compute_all_neighbors(data1, data2=None, just_nn=True):
    """
    For each point in data1, compute a ranked list of neighbor indices from data2.
    If data2 is not provided, compute neighbors relative to data1

    :param ndarray data1:
        an m1 x n dim matrix with observations on the rows
    :param ndarray data2:
        an m2 x n dim matrix with observations on the rows

    :returns ndarray:
        an m1 x m2 dim matrix with the distance-sorted indices of neighbors on the rows
    """
    from scipy.spatial.distance import cdist

    if data2 is None:
        data2 = data1

    dists = cdist(data1, data2)

    if just_nn:
        nns = np.zeros((data1.shape[0]), dtype=int)
    else:
        nns = np.zeros(dists.shape, dtype=int)

    for i in xrange(dists.shape[0]):
        if just_nn:
            nns[i] = np.argmin(dists[i])
        else:
            nns[i] = np.argsort(dists[i])

    return nns


def get_proportion_nns_with_same_coarse_codes(data, model, nns=None):
    """
    """
    N = data.shape[0]

    # Compute nearest neighbors if not provided
    if nns is None:
        nns = compute_all_neighbors(data)

    # Compute coarse codes for data
    coarse_codes = []
    for d in data:
        c = model.predict_coarse(d)
        coarse_codes.append(c)

    # Count the number of NNs that share the same coarse codes
    count = 0
    for i in xrange(N):
        nn = nns[i]
        if coarse_codes[i] == coarse_codes[nn]:
            count += 1

    return float(count) / N


def get_cell_histogram(data, model):
    # Compute cells for data
    cells = []
    for d in data:
        c = model.predict_coarse(d)
        cell = model.get_cell_id_for_coarse_codes(c)
        cells.append(cell)

    return np.histogram(cells, bins=range(model.V ** 2))[0]


def get_proportion_of_reconstructions_with_same_codes(data, model):
    N = data.shape[0]

    # Compute coarse codes for data
    count = 0
    for d in data:
        c1 = model.predict(d)
        r = model.reconstruct(c1)
        c2 = model.predict(r)
        if c1 == c2:
            count += 1

    return float(count) / N


def get_recall(searcher, queries, nns, thresholds=[1, 10, 100, 1000], normalize=True, verbose=False):
    """
    Given a LOPQSearcher object with indexed data and groundtruth nearest neighbors for a set of test
    query vectors, collect and return recall statistics.

    :param LOPQSearcher searcher:
        a searcher that contains the indexed nearest neighbors
    :param ndarray queries:
        a collect of test vectors with vectors on the rows
    :param ndarray nns:
        a list of true nearest neighbor ids for each vector in queries
    :param list thresholds:
        the recall thresholds to evaluate - the last entry defines the number of
        results to retrieve before ranking
    :param bool normalize:
        flag to indicate whether the result should be normalized by the number of queries
    :param bool verbose:
        flag to print every 50th search to visualize progress

    :return list:
        a list of recalls for each specified threshold level
    :return float:
        the elapsed query time
    """

    recall = np.zeros(len(thresholds))
    query_time = 0.0
    for i, d in enumerate(queries):

        nn = nns[i]

        start = time.clock()
        results, cells_visited = searcher.search(d, thresholds[-1])
        query_time += time.clock() - start

        if verbose and i % 50 == 0:
            print '%d cells visitied for query %d' % (cells_visited, i)

        for j, res in enumerate(results):
            rid, code = res

            if rid == nn:
                for k, t in enumerate(thresholds):
                    if j < t:
                        recall[k] += 1

    if normalize:
        N = queries.shape[0]
        return recall / N, query_time / N
    else:
        return recall, query_time


def get_subquantizer_distortion(data, model):
    from .model import compute_residuals, project_residuals_to_local

    first_half, second_half = np.split(data, 2, axis=1)

    r1, a1 = compute_residuals(first_half, model.Cs[0])
    r2, a2 = compute_residuals(second_half, model.Cs[1])

    p1 = project_residuals_to_local(r1, a1, model.Rs[0], model.mus[0])
    p2 = project_residuals_to_local(r2, a2, model.Rs[1], model.mus[1])

    pall = np.concatenate((p1, p2), axis=1)
    suball = model.subquantizers[0] + model.subquantizers[1]

    dists = np.array([sum(np.linalg.norm(compute_residuals(d, c)[0], ord=2, axis=1) ** 2) for c, d in zip(suball, np.split(pall, 8, axis=1))])

    return dists / data.shape[0]



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: lopq_model.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='lopq_model.proto',
  package='com.flickr.vision.lopq',
  serialized_pb=_b('\n\x10lopq_model.proto\x12\x16\x63om.flickr.vision.lopq\"\x1c\n\x06Vector\x12\x12\n\x06values\x18\x01 \x03(\x02\x42\x02\x10\x01\"+\n\x06Matrix\x12\x12\n\x06values\x18\x01 \x03(\x02\x42\x02\x10\x01\x12\r\n\x05shape\x18\x02 \x03(\r\"\x80\x02\n\x0fLOPQModelParams\x12\t\n\x01\x44\x18\x01 \x01(\r\x12\t\n\x01V\x18\x02 \x01(\r\x12\t\n\x01M\x18\x03 \x01(\r\x12\x19\n\x11num_subquantizers\x18\x04 \x01(\r\x12*\n\x02\x43s\x18\x05 \x03(\x0b\x32\x1e.com.flickr.vision.lopq.Matrix\x12*\n\x02Rs\x18\x06 \x03(\x0b\x32\x1e.com.flickr.vision.lopq.Matrix\x12+\n\x03mus\x18\x07 \x03(\x0b\x32\x1e.com.flickr.vision.lopq.Vector\x12,\n\x04subs\x18\x08 \x03(\x0b\x32\x1e.com.flickr.vision.lopq.MatrixB/\n\x16\x63om.flickr.vision.lopqB\x13LOPQModelParametersH\x01')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_VECTOR = _descriptor.Descriptor(
  name='Vector',
  full_name='com.flickr.vision.lopq.Vector',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='values', full_name='com.flickr.vision.lopq.Vector.values', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=44,
  serialized_end=72,
)


_MATRIX = _descriptor.Descriptor(
  name='Matrix',
  full_name='com.flickr.vision.lopq.Matrix',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='values', full_name='com.flickr.vision.lopq.Matrix.values', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))),
    _descriptor.FieldDescriptor(
      name='shape', full_name='com.flickr.vision.lopq.Matrix.shape', index=1,
      number=2, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=74,
  serialized_end=117,
)


_LOPQMODELPARAMS = _descriptor.Descriptor(
  name='LOPQModelParams',
  full_name='com.flickr.vision.lopq.LOPQModelParams',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='D', full_name='com.flickr.vision.lopq.LOPQModelParams.D', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='V', full_name='com.flickr.vision.lopq.LOPQModelParams.V', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='M', full_name='com.flickr.vision.lopq.LOPQModelParams.M', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='num_subquantizers', full_name='com.flickr.vision.lopq.LOPQModelParams.num_subquantizers', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Cs', full_name='com.flickr.vision.lopq.LOPQModelParams.Cs', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='Rs', full_name='com.flickr.vision.lopq.LOPQModelParams.Rs', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='mus', full_name='com.flickr.vision.lopq.LOPQModelParams.mus', index=6,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='subs', full_name='com.flickr.vision.lopq.LOPQModelParams.subs', index=7,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=120,
  serialized_end=376,
)

_LOPQMODELPARAMS.fields_by_name['Cs'].message_type = _MATRIX
_LOPQMODELPARAMS.fields_by_name['Rs'].message_type = _MATRIX
_LOPQMODELPARAMS.fields_by_name['mus'].message_type = _VECTOR
_LOPQMODELPARAMS.fields_by_name['subs'].message_type = _MATRIX
DESCRIPTOR.message_types_by_name['Vector'] = _VECTOR
DESCRIPTOR.message_types_by_name['Matrix'] = _MATRIX
DESCRIPTOR.message_types_by_name['LOPQModelParams'] = _LOPQMODELPARAMS

Vector = _reflection.GeneratedProtocolMessageType('Vector', (_message.Message,), dict(
  DESCRIPTOR = _VECTOR,
  __module__ = 'lopq_model_pb2'
  # @@protoc_insertion_point(class_scope:com.flickr.vision.lopq.Vector)
  ))
_sym_db.RegisterMessage(Vector)

Matrix = _reflection.GeneratedProtocolMessageType('Matrix', (_message.Message,), dict(
  DESCRIPTOR = _MATRIX,
  __module__ = 'lopq_model_pb2'
  # @@protoc_insertion_point(class_scope:com.flickr.vision.lopq.Matrix)
  ))
_sym_db.RegisterMessage(Matrix)

LOPQModelParams = _reflection.GeneratedProtocolMessageType('LOPQModelParams', (_message.Message,), dict(
  DESCRIPTOR = _LOPQMODELPARAMS,
  __module__ = 'lopq_model_pb2'
  # @@protoc_insertion_point(class_scope:com.flickr.vision.lopq.LOPQModelParams)
  ))
_sym_db.RegisterMessage(LOPQModelParams)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\n\026com.flickr.vision.lopqB\023LOPQModelParametersH\001'))
_VECTOR.fields_by_name['values'].has_options = True
_VECTOR.fields_by_name['values']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))
_MATRIX.fields_by_name['values'].has_options = True
_MATRIX.fields_by_name['values']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\020\001'))
# @@protoc_insertion_point(module_scope)



# Copyright 2015, Yahoo Inc.
# Licensed under the terms of the Apache License, Version 2.0. See the LICENSE file associated with the project for terms.
import numpy as np
import multiprocessing
from itertools import chain


def iterate_splits(x, splits):
    """
    A helper to iterate subvectors.

    :param ndarray x:
        a vector to iterate over
    :param int splits:
        the number of subvectors
    :returns (np.array, int):
        subvector, split index pairs
    """
    split_size = len(x) / splits
    for split in xrange(splits):
        start = split * split_size
        yield x[start:start + split_size], split


def concat_new_first(arrs):
    """
    Helper to concatenate a list of ndarrays along a new first dimension.
    """
    arrs = map(lambda x: x[np.newaxis, ...], arrs)
    return np.concatenate(arrs, axis=0)


def predict_cluster(x, centroids):
    """
    Given a vector of dimension D and a matrix of centroids of dimension VxD,
    return the id of the closest cluster

    :params np.array x:
        the data to assign
    :params np.array centroids:
        a matrix of cluster centroids
    :returns int:
        cluster assignment
    """
    return ((x - centroids) ** 2).sum(axis=1).argmin(axis=0)


def load_xvecs(filename, base_type='f', max_num=None):
    """
    A helper to read in sift1m binary dataset. This parses the
    binary format described at http://corpus-texmex.irisa.fr/.

    :returns ndarray:
        a N x D array, where N is the number of observations
        and D is the number of features
    """
    import os
    import struct

    format_code, format_size, py_type = {
        'f': ('f', 4, float),
        'i': ('I', 4, int),
        'b': ('B', 1, float)
    }[base_type]

    size = os.path.getsize(filename)

    f = open(filename, 'rb')
    D = np.uint32(struct.unpack('I', f.read(4))[0])
    N = size / (4 + D * format_size)

    if max_num is None:
        max_num = N

    f.seek(0)
    A = np.zeros((max_num, D), dtype=py_type)
    for i in xrange(max_num):
        for j in xrange(D + 1):
            if j == 0:
                np.uint32(struct.unpack(format_code, f.read(4)))
            else:
                A[i, j - 1] = py_type(struct.unpack(format_code, f.read(format_size))[0])
    f.close()
    return np.squeeze(A)


def save_xvecs(data, filename, base_type='f'):
    """
    A helper to save an ndarray in the binary format as is expected in
    load_xvecs above.
    """
    import struct

    format_code, format_size, py_type = {
        'f': ('f', 4, float),
        'i': ('I', 4, int),
        'b': ('B', 1, float)
    }[base_type]

    f = open(filename, 'wb')
    for d in data:

        if hasattr(d, "__len__"):
            D = len(d)

            f.write(struct.pack('<I', D))
            for x in d:
                f.write(struct.pack(format_code, x))
        else:
            D = 1
            f.write(struct.pack('<I', D))
            f.write(struct.pack(format_code, d))

    f.flush()
    f.close()


def parmap(f, X, nprocs=multiprocessing.cpu_count()):
    """
    Parallel map implementation adapted from http://stackoverflow.com/questions/3288595/multiprocessing-using-pool-map-on-a-function-defined-in-a-class
    """

    def func_wrap(f, q_in, q_out):
        while True:
            i, x = q_in.get()
            if i is None:
                break
            q_out.put((i, f(x)))

    q_in = multiprocessing.Queue(1)
    q_out = multiprocessing.Queue()

    proc = [multiprocessing.Process(target=func_wrap, args=(f, q_in, q_out)) for _ in range(nprocs)]
    for p in proc:
        p.daemon = True
        p.start()

    sent = [q_in.put((i, x)) for i, x in enumerate(X)]
    [q_in.put((None, None)) for _ in range(nprocs)]
    res = [q_out.get() for _ in range(len(sent))]

    [p.join() for p in proc]
    [p.terminate() for p in proc]

    return [x for i, x in sorted(res)]


def get_chunk_ranges(N, num_procs):
    """
    A helper that given a number N representing the size of an iterable and the num_procs over which to
    divide the data return a list of (start_index, end_index) pairs that divide the data as evenly as possible
    into num_procs buckets.
    """
    per_thread = N / num_procs
    allocation = [per_thread] * num_procs
    allocation[0] += N - num_procs * per_thread
    data_ranges = [0] + reduce(lambda acc, num: acc + [num + (acc[-1] if len(acc) else 0)], allocation, [])
    data_ranges = [(data_ranges[i], data_ranges[i + 1]) for i in range(len(data_ranges) - 1)]
    return data_ranges


def compute_codes_parallel(data, model, num_procs=4):
    """
    A helper function that parallelizes the computation of LOPQ codes in 
    a configurable number of processes.

    :param ndarray data:
        an ndarray of data points
    :param LOPQModel model:
        a model instance to use to compute codes
    :param int num_procs:
        the number of processes to spawn

    :returns iterable:
        an iterable of computed codes in the input order
    """
    def compute_partition(data):
        return [model.predict(d) for d in data]

    N = len(data)
    partitions = [data[a:b] for a, b in get_chunk_ranges(N, num_procs)]
    codes = parmap(compute_partition, partitions, num_procs)

    return chain(*codes)



#!/usr/bin/env python

# Copyright 2015, Yahoo Inc.
# Licensed under the terms of the Apache License, Version 2.0. See the LICENSE file associated with the project for terms.
import os
import json
from setuptools import setup


# Package Metadata filename
METADATA_FILENAME = 'lopq/package_metadata.json'
BASEPATH = os.path.dirname(os.path.abspath(__file__))


# Long description of package
LONG_DESCRIPTION = """
# Locally Optimized Product Quantization

This is Python training and testing code for Locally Optimized Product Quantization (LOPQ) models, as well as Spark scripts to scale training to hundreds of millions of vectors. The resulting model can be used in Python with code provided here or deployed via a Protobuf format to, e.g., search backends for high performance approximate nearest neighbor search.

### Overview

Locally Optimized Product Quantization (LOPQ) [1] is a hierarchical quantization algorithm that produces codes of configurable length for data points. These codes are efficient representations of the original vector and can be used in a variety of ways depending on application, including as hashes that preserve locality, as a compressed vector from which an approximate vector in the data space can be reconstructed, and as a representation from which to compute an approximation of the Euclidean distance between points.

Conceptually, the LOPQ quantization process can be broken into 4 phases. The training process also fits these phases to the data in the same order.

1. The raw data vector is PCA'd to `D` dimensions (possibly the original dimensionality). This allows subsequent quantization to more efficiently represent the variation present in the data.
2. The PCA'd data is then product quantized [2] by two k-means quantizers. This means that each vector is split into two subvectors each of dimension `D / 2`, and each of the two subspaces is quantized independently with a vocabulary of size `V`. Since the two quantizations occur independently, the dimensions of the vectors are permuted such that the total variance in each of the two subspaces is approximately equal, which allows the two vocabularies to be equally important in terms of capturing the total variance of the data. This results in a pair of cluster ids that we refer to as "coarse codes".
3. The residuals of the data after coarse quantization are computed. The residuals are then locally projected independently for each coarse cluster. This projection is another application of PCA and dimension permutation on the residuals and it is "local" in the sense that there is a different projection for each cluster in each of the two coarse vocabularies. These local rotations make the next and final step, another application of product quantization, very efficient in capturing the variance of the residuals.
4. The locally projected data is then product quantized a final time by `M` subquantizers, resulting in `M` "fine codes". Usually the vocabulary for each of these subquantizers will be a power of 2 for effective storage in a search index. With vocabularies of size 256, the fine codes for each indexed vector will require `M` bytes to store in the index.

The final LOPQ code for a vector is a `(coarse codes, fine codes)` pair, e.g. `((3, 2), (14, 164, 83, 49, 185, 29, 196, 250))`.

### Nearest Neighbor Search

A nearest neighbor index can be built from these LOPQ codes by indexing each document into its corresponding coarse code bucket. That is, each pair of coarse codes (which we refer to as a "cell") will index a bucket of the vectors quantizing to that cell.

At query time, an incoming query vector undergoes substantially the same process. First, the query is split into coarse subvectors and the distance to each coarse centroid is computed. These distances can be used to efficiently compute a priority-ordered sequence of cells [3] such that cells later in the sequence are less likely to have near neighbors of the query than earlier cells. The items in cell buckets are retrieved in this order until some desired quota has been met.

After this retrieval phase, the fine codes are used to rank by approximate Euclidean distance. The query is projected into each local space and the distance to each indexed item is estimated as the sum of the squared distances of the query subvectors to the corresponding subquantizer centroid indexed by the fine codes.

NN search with LOPQ is highly scalable and has excellent properties in terms of both index storage requirements and query-time latencies when implemented well.

#### References

For more information and performance benchmarks can be found at http://image.ntua.gr/iva/research/lopq/.

1. Y. Kalantidis, Y. Avrithis. [Locally Optimized Product Quantization for Approximate Nearest Neighbor Search.](http://image.ntua.gr/iva/files/lopq.pdf) CVPR 2014.
2. H. Jegou, M. Douze, and C. Schmid. [Product quantization for nearest neighbor search.](https://lear.inrialpes.fr/pubs/2011/JDS11/jegou_searching_with_quantization.pdf) PAMI, 33(1), 2011.
3. A. Babenko and V. Lempitsky. [The inverted multi-index.](http://www.computer.org/csdl/trans/tp/preprint/06915715.pdf) CVPR 2012.
"""

# Create a dictionary of our arguments, this way this script can be imported
# without running setup() to allow external scripts to see the setup settings.
setup_arguments = {
    'name': 'lopq',
    'version': '1.0.0',
    'author': 'Clayton Mellina,Yannis Kalantidis,Huy Nguyen',
    'author_email': 'clayton@yahoo-inc.com',
    'url': 'http://github.com/yahoo/lopq',
    'license': 'Apache-2.0',
    'keywords': ['lopq', 'locally optimized product quantization', 'product quantization', 'compression', 'ann', 'approximate nearest neighbor', 'similarity search'],
    'packages': ['lopq'],
    'long_description': LONG_DESCRIPTION,
    'description': 'Python code for training and deploying Locally Optimized Product Quantization (LOPQ) for approximate nearest neighbor search of high dimensional data.',
    'classifiers': [
            'Development Status :: 5 - Production/Stable',
            'License :: OSI Approved :: Apache Software License',
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'Natural Language :: English',
            'Operating System :: MacOS :: MacOS X',
            'Operating System :: Microsoft :: Windows',
            'Operating System :: POSIX',
            'Operating System :: POSIX',
            'Operating System :: Unix',
            'Programming Language :: Python :: 2.7',
            'Topic :: Scientific/Engineering',
            'Topic :: Software Development'
    ],
    'package_data': {
        'lopq': ['package_metadata.json']
    },
    'platforms': 'Windows,Linux,Solaris,Mac OS-X,Unix',
    'include_package_data': True,
    'install_requires': ['protobuf>=2.6', 'numpy>=1.9', 'scipy>=0.14', 'scikit-learn>=0.15', 'lmdb>=0.87']
}


class Git(object):
    """
    Simple wrapper class to the git command line tools
    """
    version_list = ['0', '7', '0']

    def __init__(self, version=None):
        if version:
            self.version_list = version.split('.')

    @property
    def version(self):
        """
        Generate a Unique version value from the git information
        :return:
        """
        git_rev = len(os.popen('git rev-list HEAD').readlines())
        if git_rev != 0:
            self.version_list[-1] = '%d' % git_rev
        version = '.'.join(self.version_list)
        return version

    @property
    def branch(self):
        """
        Get the current git branch
        :return:
        """
        return os.popen('git rev-parse --abbrev-ref HEAD').read().strip()

    @property
    def hash(self):
        """
        Return the git hash for the current build
        :return:
        """
        return os.popen('git rev-parse HEAD').read().strip()

    @property
    def origin(self):
        """
        Return the fetch url for the git origin
        :return:
        """
        for item in os.popen('git remote -v'):
            split_item = item.strip().split()
            if split_item[0] == 'origin' and split_item[-1] == '(push)':
                return split_item[1]


def add_scripts_to_package():
    """
    Update the "scripts" parameter of the setup_arguments with any scripts
    found in the "scripts" directory.
    :return:
    """
    global setup_arguments

    if os.path.isdir('scripts'):
        setup_arguments['scripts'] = [
            os.path.join('scripts', f) for f in os.listdir('scripts')
        ]


def get_and_update_package_metadata():
    """
    Update the package metadata for this package if we are building the package.
    :return:metadata - Dictionary of metadata information
    """
    global setup_arguments
    global METADATA_FILENAME

    if not os.path.exists('.git') and os.path.exists(METADATA_FILENAME):
        with open(METADATA_FILENAME) as fh:
            metadata = json.load(fh)
    else:
        git = Git(version=setup_arguments['version'])
        metadata = {
            'version': git.version,
            'git_hash': git.hash,
            'git_origin': git.origin,
            'git_branch': git.branch
        }
        with open(METADATA_FILENAME, 'w') as fh:
            json.dump(metadata, fh)
    return metadata


if __name__ == '__main__':
    # We're being run from the command line so call setup with our arguments
    metadata = get_and_update_package_metadata()
    setup_arguments['version'] = metadata['version']
    setup(**setup_arguments)



# Copyright 2015, Yahoo Inc.
# Licensed under the terms of the Apache License, Version 2.0. See the LICENSE file associated with the project for terms.
import model
import search
import utils
from .model import LOPQModel
from .search import LOPQSearcher, multisequence

__all__ = [LOPQModel, LOPQSearcher, multisequence, model, search, utils]
